#!/usr/bin/env python

# Copyright (c) 2007, 2013 J. Lewis Muir <jlmuir@imca-cat.org>
#
# Permission to use, copy, modify, and distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

#
# For usage, see printUsage function or run with -h option.
#

import getopt
import asynchat
import asyncore
import os
import re
import signal
import socket
import sys
import time

PORT = 50433
IS_DEBUG_OUTPUT_ENABLED = False

def main(args):
  parseCommandLineOptions(args)
  ConnectionDispatcher(PORT)
  asyncore.loop()

def getProgramName(args=None):
  if args == None: args = sys.argv
  if len(args) == 0 or args[0] == "-c": return "PROGRAM_NAME"
  return os.path.basename(args[0])

def initializeSignalHandlers():
  signalToNameMap = {
    signal.SIGHUP:"SIGHUP",
    signal.SIGINT:"SIGINT",
    signal.SIGPIPE:"SIGPIPE",
    signal.SIGALRM:"SIGALRM",
    signal.SIGTERM:"SIGTERM",
    signal.SIGUSR1:"SIGUSR1",
    signal.SIGUSR2:"SIGUSR2"
  }

  def signalHandler(signum, frame):
    if signum in [signal.SIGINT, signal.SIGTERM]:
      print
      print "%s caught. Exiting." % signalToNameMap[signum]
      sys.exit(0)
    else:
      print >> sys.stderr
      print >> sys.stderr, "%s caught. Exiting." % signalToNameMap[signum]
      sys.exit(1)

  for each in signalToNameMap.keys():
    signal.signal(each, signalHandler)

def parseCommandLineOptions(args):
  (options, extra) = getopt.getopt(args[1:], "dp:h", ["debug", "port=",
    "help"])

  if len(extra) > 0:
    print >> sys.stderr, "Error: unexpected command line option \"%s\"" % \
      extra[0]
    printUsage(sys.stderr)
    sys.exit(1)

  for eachOptName, eachOptValue in options:
    if eachOptName in ("-d", "--debug"):
      global IS_DEBUG_OUTPUT_ENABLED
      IS_DEBUG_OUTPUT_ENABLED = True
    elif eachOptName in ("-p", "--port"):
      global PORT
      PORT = int(eachOptValue)
    elif eachOptName in ("-h", "--help"):
      printUsage(sys.stdout)
      sys.exit(0)

def printUsage(outStream):
  print >> outStream, """\
Usage: %s [OPTION]...
Options:
  -d,--debug                Print debug messages to stderr
  -p,--port=NUMBER          Listen on the specified port NUMBER for incoming
                            connections (default: %d)
  -h,--help                 Print usage message and exit
""" % (getProgramName(), PORT)

class ConnectionDispatcher(asyncore.dispatcher):
  def __init__(self, port):
    asyncore.dispatcher.__init__ (self)
    self.port = port
    self.create_socket(socket.AF_INET, socket.SOCK_STREAM)
    self.set_reuse_addr()
    self.bind(("", port))
    self.listen(5)

  def handle_accept(self):
    ConnectionSession(self.accept())

class ConnectionSession(asynchat.async_chat):
  def __init__(self, (conn, addr)):
    asynchat.async_chat.__init__(self, conn)
    self.set_terminator("\r\n")
    self.outputTerminator = "\r\n"
    self.buffer = ""

    self.levels = [50.0, 50.0, 50.0, 50.0]
    self.invalidLevels = [None, None, None, None]
    self.levelGetCmdRe = re.compile(r"^CH(?P<channel>[1-4]):LEV(?:EL)?\?$")
    self.levelSetCmdRe = re.compile(
      r"^CH(?P<channel>[1-4]):LEV(?:EL)=(?P<value>[0-9]+\.[0-9]+)$")
    self.invalidLevelSetCmdRe = re.compile(
      r"^CH(?P<channel>[1-4]):INVALID_LEV(?:EL)=(?P<value>.*)$")

  def collect_incoming_data(self, data):
    self.buffer = self.buffer + data

  def found_terminator(self):
    data = self.buffer
    self.buffer = ""
    self.handleClientRequest(data)

  def handleClientRequest(self, request):
    request = request.strip()
    if IS_DEBUG_OUTPUT_ENABLED: sys.stderr.write("< \"%s\"\n" % request)
    if self.levelGetCmdRe.match(request) != None:
      match = self.levelGetCmdRe.match(request)
      channel = int(match.group("channel")) - 1
      if self.invalidLevels[channel] != None:
        self.sendClientResponse(self.invalidLevels[channel])
      else:
        self.sendClientResponse("%.1f" % self.levels[channel])
    elif self.levelSetCmdRe.match(request) != None:
      # internal simulation-only command
      match = self.levelSetCmdRe.match(request)
      channel = int(match.group("channel")) - 1
      self.levels[channel] = float(match.group("value"))
      self.sendClientResponse()
    elif self.invalidLevelSetCmdRe.match(request) != None:
      # internal simulation-only command
      match = self.invalidLevelSetCmdRe.match(request)
      channel = int(match.group("channel")) - 1
      self.invalidLevels[channel] = match.group("value")
      if self.invalidLevels[channel] == "nil":
        self.invalidLevels[channel] = None
      self.sendClientResponse()
    else:
      self.sendClientResponse("-101, \"Unrecognized command\"")

  def sendClientResponse(self, response=""):
    self.push(response + self.outputTerminator)
    if IS_DEBUG_OUTPUT_ENABLED: sys.stderr.write("> \"%s\"\n" % response)

if __name__ == '__main__':
  try:
    initializeSignalHandlers()
    main(sys.argv)
  except Exception, e:
    if isinstance(e, SystemExit):
      raise e
    else:
      print >> sys.stderr
      print >> sys.stderr, "Error: %s" % e
      sys.exit(1)
